#
# Copyright (c) 2020-2021 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
import decimal
import json

from .register_row import RegisterRow

def sensor_port_name(index):
    '''Sensor port name for sensor port with index INDEX, which must be
    in the range from 0..SENSOR_PORT_COUNT-1.

    '''
    return 'S%u' % (index + 1)

class Local:
    '''Fetch data from /api/local.'''

    # Sensor parameter spec strings:
    #   For l=...:
    SPEC_L1 = 'L1'
    SPEC_L2 = 'L2'
    SPEC_L3 = 'L3'
    SPEC_L1_L2 = 'L12'
    SPEC_L2_L3 = 'L23'
    SPEC_L3_L1 = 'L31'
    SPEC_LDC = 'Ldc'

    # Sensor names as they appear in the output:
    NAME_L1 = 'L1'
    NAME_L2 = 'L2'
    NAME_L3 = 'L3'
    NAME_L1_L2 = 'L1-L2'
    NAME_L2_L3 = 'L2-L3'
    NAME_L3_L1 = 'L3-L1'
    NAME_LDC = 'Ldc'
    NAME_TEMP_PCB = 'Tpcb'
    NAME_HUMID_PCB = 'Hpcb'

    METRIC_RATE = 'rate'
    METRIC_CUMUL = 'cumul'
    METRIC_TYPE = 'type'

    MEASUREMENT_NORMAL = 'n'
    MEASUREMENT_MEAN = 'm'
    MEASUREMENT_FREQ = 'f'

    SECTION_VALUES = 'values'
    SECTION_ENERGY = 'energy'
    SECTION_APPARENT = 'apparent'

    def __init__(self, dev, params, **kwargs):
        self.raw = dev.get('/local', params=params, **kwargs)

    def ts(self):
        '''Return the timestamp of the local data as a Decimal().'''
        if self.raw is None:
            return None
        return decimal.Decimal(self.raw['ts'])

    def type_code(self, sensor_name):
        '''Return the type-code of sensor SENSOR_NAME.'''
        if self.raw is None:
            return None
        try:
            return self.raw['values'][sensor_name]['type']
        except KeyError:
            pass
        return None

    def rate(self, sensor_name, measurement=MEASUREMENT_NORMAL,
             section=SECTION_VALUES):
        '''Return the rate of change value of sensor SENSOR_NAME.  By default,
        the normal value for the sensor is returned.  The mean value
        or frequency can be requested by passing MEASUREMENT_MEAN or
        MEASUREMENT_FREQ, respectively, as MEASUREMENT.  If the value
        is not available, None is returned.

        '''
        if self.raw is None:
            return None
        try:
            m = self.raw[section][sensor_name]['rate']
            if section == Local.SECTION_VALUES:
                return m[measurement]
            return m
        except KeyError:
            pass
        return None

    def cumul(self, sensor_name, measurement=MEASUREMENT_NORMAL,
              section=SECTION_VALUES):
        '''Return the cumulative value of sensor SENSOR_NAME.  By default, the
        normal value for the sensor is returned.  The mean value or
        frequency can be requested by passing MEASUREMENT_MEAN or
        MEASUREMENT_FREQ, respectively, as MEASUREMENT.  If the value
        is not available, None is returned.

        '''
        if self.raw is None:
            return None
        try:
            m = self.raw[section][sensor_name]['cumul']
            if section == Local.SECTION_VALUES:
                return int(m[measurement])
            return int(m)
        except KeyError:
            pass
        return None

    def row(self, metric=METRIC_CUMUL, measurement=MEASUREMENT_NORMAL,
            section=SECTION_VALUES):
        '''Return a timestamped row of register data.  By default, the
        cumulative normal values of section "values" is returned.
        Non-default rows can be selected by passing the desired values
        for METRIC, MEASUREMENT, and/or SECTION.

        '''
        if self.raw is None:
            return None
        section = self.raw[section]
        regs = {}
        for key, metrics in section.items():
            if metric == Local.METRIC_TYPE:
                val = metrics[metric]
            else:
                val = metrics[metric]
                if section == Local.SECTION_VALUES:
                    val = val[measurement]
                if metric == Local.METRIC_CUMUL:
                    val = int(val)	# convert from decimal string to integer
            regs[key] = val
        return RegisterRow(self.ts(), regs)

    def __str__(self):
        return json.dumps(self.raw)



#
#   Copyright (c) 2014, 2016-2017, 2020 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
#   All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""Basic terminal emulation for QT."""

import re

from PySide2.QtWidgets import QApplication
from PySide2.QtGui import QCursor, QTextCursor

from . import ansi2html

CLEAR_LINE_PATTERN = re.compile(r'.*\033\[2K')
CURSOR_REPORT_PATTERN = re.compile(r'\033\[6n')
# ANSI CSI sequence: ESC [ followed by any number of parameter bytes in
# range 0x30-0x3f, followed by any number of intermediate bytes in
# range 0x20-0x2f, followed by a single final byte in range 0x40-0x74:
INCOMPLETE_ANSI_CSI_PATTERN = re.compile(r'\033(\[([0-?]*[ -/]*)?)?')

def split_keepends(string, pattern):
    res = []
    while len(string) > 0:
        m = re.search(pattern, string)
        if m:
            last_matched = m.end(0)
            res.append(string[0:last_matched])
            string = string[last_matched:]
        else:
            res.append(string)
            break
    return res

def incomplete_ansi_csi(string):
    '''Return True if the string forms a partial ANSI CSI ESCape sequence,
    False otherwise.

    '''
    m = INCOMPLETE_ANSI_CSI_PATTERN.match(string)
    if m is None:
        return False
    return m.end() == len(string)

class Terminal:
    """Provides basic terminal emulation for a QT plain text widget."""

    def __init__(self, plain_text_edit):
        self.plain_text_edit = plain_text_edit
        self.partial_line = ''

    def write(self, string):
        """Write STRING to the terminal.  The string may consist of multiple
        lines.

        """

        if len(string) == 0:
            return

        string = self.partial_line + string
        self.partial_line = ''

        # if the string ends with an incomplete ANSI CSI sequence,
        # hold off processing the partial escape sequence until it's
        # complete but process the text up to that point immediately:
        for m in re.finditer(r'\033', string):
            if incomplete_ansi_csi(string[m.start():]):
                self.partial_line = string[m.start():]
                string = string[0:m.start()]
                break
        if len(self.partial_line) == 0 and string[-1] == '\r':
            # if the string ends with a carriage-return, hold off
            # processing the carriage return until we get the next
            # character:
            self.partial_line = '\r'
            string = string[:-1]

        lines = split_keepends(string, re.compile('(\r\n|\n|\r|\b)'))
        for line in lines:
            # position cursor at end of text (in case someone clicked in
            # the middle of the text):
            cursor = self.plain_text_edit.textCursor()
            cursor.movePosition(QTextCursor.End)
            self.plain_text_edit.setTextCursor(cursor)

            back_space = False
            line_feed = False
            txt = line
            if len(txt) >= 1:
                if txt[-1] == '\b':
                    back_space = True
                    txt = txt[:-1]
                elif txt[-1] == '\r':
                    txt = ''
                    cursor.movePosition(QTextCursor.StartOfBlock,
                                        QTextCursor.KeepAnchor)
                    cursor.removeSelectedText()
                elif len(txt) >= 2 and txt[-2] == '\r' and txt[-1] == '\n':
                    txt = txt[:-2]
                    line_feed = True
                elif txt[-1] == '\n':
                    txt = txt[:-1]
                    line_feed = True
            txt = CURSOR_REPORT_PATTERN.sub('', txt)
            m = CLEAR_LINE_PATTERN.match(txt)
            if m:
                txt = txt[m.end():]
                cursor.movePosition(QTextCursor.StartOfBlock)
                cursor.movePosition(QTextCursor.EndOfBlock,
                                    QTextCursor.KeepAnchor)
                cursor.removeSelectedText()

            if len(txt) > 0:
                # Unfortunately, the "<pre>" seems to for the font to
                # "Courier New" but without "<pre>" whitespace is not
                # preserved by insertHtml.  Argh...
                # Use qt5ct to configure the "Fixed width" font which will
                # be used for <pre>.
                html = '<pre>' + ansi2html.convert(txt) + '</pre>'
                cursor.insertHtml(html)
            if back_space:
                cursor.deletePreviousChar()
                QApplication.processEvents()
            elif line_feed:
                # Each appendHtml() writes a separate line.  Newline
                # characters are ignored.
                self.plain_text_edit.appendHtml('')
        self.plain_text_edit.centerCursor()

    def flush(self):
        """Flush pending output."""

def test_incomplete_ansi_csi(seq):
    print('Testing: %s' % seq)
    CSI = '\033['
    csi_seq = CSI + seq
    for l in range(len(csi_seq) - 1):
        if not incomplete_ansi_csi(csi_seq[0:l+1]):
            print('Error: CSI test sequence %s, l=%d returned False' % (seq, l))
    if incomplete_ansi_csi(csi_seq):
        print('Error: CSI test sequence %s returned True' % seq)

if __name__ == '__main__':
    if incomplete_ansi_csi('hi there'):
        print('Error: CSI test sequence "hi there" returned True')
    if incomplete_ansi_csi('x\033[s'):
        print('Error: CSI test sequence "xCSIs" returned True')
    test_incomplete_ansi_csi('6n')
    test_incomplete_ansi_csi('16;34H')	# set cursor position
    test_incomplete_ansi_csi('s')	# save cursor position

    import sys
    from PySide2 import QtCore, QtWidgets
    from PySide2.QtCore import Qt, QBasicTimer
    from PySide2.QtWidgets import QApplication, QMainWindow, QMessageBox

    class Ui_MainWindow:
        def setupUi(self, MainWindow):
            MainWindow.setObjectName('Terminal Test')
            MainWindow.resize(908, 480)
            self.centralwidget = QtWidgets.QWidget(MainWindow)
            self.centralwidget.setObjectName("centralwidget")
            self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
            self.verticalLayout.setObjectName("verticalLayout")
            self.plainTextEdit = QtWidgets.QPlainTextEdit(self.centralwidget)
            self.plainTextEdit.setObjectName("plainTextEdit")
            self.plainTextEdit \
                .viewport().setProperty('cursor',
                                        QCursor(QtCore.Qt.IBeamCursor))
            self.plainTextEdit.setMouseTracking(False)
            self.plainTextEdit.setFocusPolicy(QtCore.Qt.NoFocus)
            self.plainTextEdit.setAcceptDrops(False)
            self.plainTextEdit.setStyleSheet('background-color: black;\n'
                                             'color: white;\n')
            self.plainTextEdit.setLineWidth(1)
            self.plainTextEdit.setUndoRedoEnabled(False)
            self.plainTextEdit.setLineWrapMode(QtWidgets.QPlainTextEdit.WidgetWidth)
            self.plainTextEdit.setReadOnly(True)
            self.plainTextEdit.setTabStopWidth(80)
            self.plainTextEdit.setCursorWidth(16)
            self.plainTextEdit.setBackgroundVisible(False)
            self.plainTextEdit.setCenterOnScroll(False)
            self.verticalLayout.addWidget(self.plainTextEdit)
            MainWindow.setCentralWidget(self.centralwidget)

            self.retranslateUi(MainWindow)
            QtCore.QMetaObject.connectSlotsByName(MainWindow)

        def retranslateUi(self, MainWindow):
            pass

    class UI(QMainWindow, Ui_MainWindow):
        def __init__(self):
            QMainWindow.__init__(self)
            self.setupUi(window)

            self.console = Terminal(self.plainTextEdit)
            self.console.write('\033')
            self.console.write('[')
            self.console.write('6')
            self.console.write('n')
            self.console.write('\033[6nHi There!\033[6n\n')

            self.console.write('        \t8 blanks and a tab before me\n')
            self.console.write('8 blanks and a tab after me>        \t')
            self.console.write('<\n')
            self.console.write('8 blanks        \tand a tab\n')

            self.console.write('GOT mo')
            self.console.write('del\r')
            self.console.write('\n')


            self.console.write('Feeling')
            self.console.write(' a ')
            self.console.write('little \033[44mblue\033[0m?\r\n')

            self.console.write('80 columns:\n')
            self.console.write('*'*80 + '\n')

            self.console.write('132 columns:\n')
            self.console.write('*'*132 + '\n')

            self.console.write('partial line')
            self.console.write('\nthis should be on a new line\n')

            # test spinner:
            self.console.write('Wait a little: x\bX')

            self.spinner_pos = 0
            self.timer = QBasicTimer()
            self.timer.start(100, self)

        def timerEvent(self, event):
            if event.timerId() != self.timer.timerId():
                super(UI, self).timerEvent(event)

            self.console.write('\b%c' % '-\\|/'[self.spinner_pos % 4])
            self.spinner_pos += 1
            if self.spinner_pos >= 16:
                self.timer.stop()
                self.console.write('\bdone with that...\n')

    app = QApplication(sys.argv)
    window = QMainWindow()
    ui = UI()
    window.show()
    sys.exit(app.exec_())



#
#   Copyright (c) 2013, 2016-2017, 2020 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""Convert terminal output with (some) ANSI escape sequences to HTML.
   Also escape any HTML special characters.

   At this time, most attribute ANSI escape sequences are converted, e.g.:

     ESC [ 0 m:         Reset attributes.
     ESC [ 41 m:        Background-color red.
     ESC [ 32 m:        Foreground-color red.
   """
import re

import html

HTML_STYLE = {
    1:  'font-weight:bold',
    3:  'font-style:italic',
    4:  'text-decoration:underline',
    9:  'text-decoration:line-through',
    22: 'font-weight:normal',
    23: 'font-style:normal',
    24: 'text-decoration:none',
    29: 'text-decoration:none',
    30: 'color:black',
    31: 'color:red',
    32: 'color:green',
    33: 'color:yellow',
    34: 'color:blue',
    35: 'color:magenta',
    36: 'color:cyan',
    37: 'color:white',
    40: 'background-color:black',
    41: 'background-color:red',
    42: 'background-color:green',
    43: 'background-color:yellow',
    44: 'background-color:blue',
    45: 'background-color:magenta',
    46: 'background-color:cyan',
    47: 'background-color:white',
}

def convert(msg):
    msg = html.escape(msg, quote=False)
    result = ''

    pattern = re.compile(r'\033\[(\d+)(;(\d+))*m')
    span_count = 0
    while True:
        m = pattern.search(msg)
        if not m:
            result += msg
            break
        result += msg[0:m.start()]
        code = int(m.group(1))
        if code == 0:
            result += span_count*'</span>'
            span_count = 0
        elif code in HTML_STYLE:
            result += '<span style="%s">' % HTML_STYLE[code]
            span_count += 1
        msg = msg[m.end(0):]
    return result + span_count*'</span>'

def selftest():
    test = 'This is in \033[32mgreen\033[0m foreground.'
    print('convert(\'%s\') -> \'%s\'' % (test, convert(test)))
    test = 'This is in \033[41mred\033[0m background.  This \033[41mtoo\033[0m.'
    print('convert(\'%s\') -> \'%s\'' % (test, convert(test)))
    test = '\033[1mBold and \033[3mitalic and \033[4munderline\033[0m and none.'
    print('convert(\'%s\') -> \'%s\'' % (test, convert(test)))

if __name__ == '__main__':
    selftest()



#
# Copyright (c) 2018-2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#  All rights reserved.
#
#  This code is the property of eGauge Systems LLC and may not be
#  copied, modified, or disclosed without any prior and written
#  permission from eGauge Systems LLC.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
import argparse
import sys

from intelhex import IntelHex

import egauge.ctid as CTid

CTid_table_addr = 0x3c0		# table goes in last 64 bytes

MAX_SERNO = 0xffffff

def is_abbrev(abbrev, full):
    if not abbrev:
        return False
    if len(abbrev) > len(full):
        return False
    return abbrev == full[0:len(abbrev)]

def auto_int(x):
    '''Accept base-prefixed decimal.'''
    return int(x, 0)

def edge_mask(x):
    try:
        return int(x)
    except ValueError:
        pass
    if is_abbrev(x, 'rising'):	# rising
        return 0x1
    if is_abbrev(x, 'falling'):
        return 0x2
    if is_abbrev(x, 'both'):
        return 0x3
    raise ValueError('Edge mask must be one of "rising", "falling", or "both".')

def cal_table_entry(string):
    pair = string.split(':')
    if len(pair) != 2:
        raise argparse.ArgumentTypeError('missing `:\'')
    idx = float(pair[0])
    if idx not in table.cal_table.keys():
        allowed = ', '.join(('%g' % k) for k in table.cal_table)
        raise argparse.ArgumentTypeError('Invalid index %s (must be one '
                                         'of %s).' % (idx, allowed))
    val = pair[1].split('/')
    if len(val) != 2:
        raise argparse.ArgumentTypeError('missing `/\'')
    v_adj = float(val[0])
    p_adj = float(val[1])
    return (idx, v_adj, p_adj)

def hexdump(title, data):
    print(title + ':', end='')
    c = 0
    for b in data:
        if c % 8 == 0:
            print('\n 0x%08x:' % c, end='')
        print(' 0x%02x' % b, end='')
        c += 1
    print('')

def encode_model_str(model_str):
    val = model_str.encode('utf-8')
    if len(val) > 4:
        print('%s: model name `%s\' is %u bytes long in UTF-8 '
              'but only up to 4 bytes are allowed.' %
              (parser.prog, model_str, len(val)), file=sys.stderr)
        sys.exit(1)
    while len(val) < 4:
        val += b'\0'
    return val

def main():
    global parser, table
    table = CTid.Table()

    parser = argparse.ArgumentParser(description=
                                     'Encode CTid parameters.',
                                     formatter_class=\
                                     argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('program_template',
                        help='Filename of the program template.', nargs=1,
                        default=None)
    parser.add_argument('-d', '--debug', nargs=1, type=int, default=[0],
                        help='Debug level')
    parser.add_argument('-o', '--output-file', nargs=1,
                        type=argparse.FileType('w'), default=[sys.stdout])
    #
    # Common information:
    #
    parser.add_argument('-S', '--sensor-type', nargs=1, default=['AC'],
                        choices=CTid.SENSOR_TYPE_NAME,
                        help='Sensor-type.')
    parser.add_argument('-M', '--manufacturer', nargs=1, default=['0'],
                        help='Manufacturer name or ID.')
    parser.add_argument('-m', '--model', nargs=1, default=[''],
                        help='Model name (up to 4 characters).')
    parser.add_argument('-n', '--serial-number', nargs=1, type=auto_int,
                        default=[0], help='Serial number.')
    parser.add_argument('-r', '--r-source', nargs=1, type=float, default=[0],
                        help='Source resistance in Ohms.')
    parser.add_argument('-l', '--r-load', nargs=1, type=float, default=[0],
                        help='Rated load resistance in Ohms.')
    parser.add_argument('-V', '--table-version', nargs=1, type=int,
                        default=[CTid.CTID_VERSION],
                        help='Version of the CTid table to generate.')
    #
    # CT Parameters:
    #
    parser.add_argument('-a', '--calibration-entry', nargs=1,
                        type=cal_table_entry, action='append',
                        help='Calibration-entry of the form X:VA/PA, where '
                        'X is the percentage of the rated current to which the '
                        'entry applies, VA is the voltage-adjustment in %% and '
                        'PA is the phase-adjustment in \u00b0.')
    parser.add_argument('-c', '--rated-current', nargs=1, type=float,
                        default=[100], help='Rated current of CT in Ampère.')
    parser.add_argument('-i', '--manufacturer-info', nargs=1, type=auto_int,
                        default=[0],
                        help='Manufacturer-specific information.')
    parser.add_argument('-p', '--phase', nargs=1, type=float, default=[0],
                        help='Phase at rated current in \u00b0.')
    parser.add_argument('-s', '--size', nargs=1, type=float, default=[0],
                        help='Size of the sensor in millimeter.')
    parser.add_argument('-t', '--voltage-temp-coeff', nargs=1, type=float,
                        default=[0],
                        help='Voltage temperature coefficient in ppm/\u00b0C.')
    parser.add_argument('-T', '--phase-temp-coeff', nargs=1, type=float,
                        default=[0],
                        help='Phase temperature coefficient in m\u00b0/\u00b0C.')
    parser.add_argument('-v', '--rated-voltage', nargs=1, type=float,
                        default=[1/3.], help='Rated voltage of CT in Volt.')
    parser.add_argument('-b', '--bias-voltage', type=float,
                        default=0, help='Output voltage when no current '
                        'is flowing through the CD.')
    #
    # Linear Parameters:
    #
    parser.add_argument('--scale', type=float, default=None, help='Scale.')
    parser.add_argument('--offset', type=float, default=None, help='Offset.')
    parser.add_argument('--delay', type=float, default=None,
                        help='Signal delay in μs.')
    parser.add_argument('--unit', type=int, default=0,
                        help='Physical unit code (0..%d).'
                        % (len(CTid.SENSOR_UNITS) - 1))
    #
    # NTC Temperature Parameters:
    #
    parser.add_argument('--ntc-a', type=float, default=None,
                        help='NTC A coefficient.')
    parser.add_argument('--ntc-b', type=float, default=None,
                        help='NTC B coefficient.')
    parser.add_argument('--ntc-c', type=float, default=None,
                        help='NTC C coefficient.')
    parser.add_argument('--ntc-m', type=float, default=None,
                        help='NTC parameter M.')
    parser.add_argument('--ntc-n', type=float, default=None,
                        help='NTC parameter N.')
    parser.add_argument('--ntc-k', type=float, default=None,
                        help='NTC parameter K.')
    #
    # Pulse Temperature Parameters:
    #
    parser.add_argument('--threshold', type=float, default=None,
                        help='Pulse threshold in Volts.')
    parser.add_argument('--hysteresis', type=float, default=None,
                        help='Hysteresis for threhold in Volts.')
    parser.add_argument('--debounce-time', type=float, default=None,
                        help='Debounce time for pulse in milli-seconds.')
    parser.add_argument('--edge-mask', type=edge_mask, default=None,
                        help='Edge-mask.  May be an integer or one of '
                        '"rising", "falling", or "both".')

    args = parser.parse_args()

    table_version = args.table_version[0]

    table.mfg_id = CTid.get_mfg_id_for_name(args.manufacturer[0])
    table.model = args.model[0]
    table.size = args.size[0]
    table.serial_number = args.serial_number[0]
    table.sensor_type = CTid.get_sensor_type_id(args.sensor_type[0])
    table.r_source = args.r_source[0]
    table.r_load = args.r_load[0]
    # CT parameters:
    table.rated_current = args.rated_current[0]
    table.voltage_at_rated_current = args.rated_voltage[0]
    table.bias_voltage = args.bias_voltage
    table.phase_at_rated_current = args.phase[0]
    table.voltage_temp_coeff = args.voltage_temp_coeff[0]
    table.phase_temp_coeff = args.phase_temp_coeff[0]
    if args.calibration_entry is not None:
        for cal in args.calibration_entry:
            c = cal[0]
            idx = c[0]
            table.cal_table[idx][0] = c[1]
            table.cal_table[idx][1] = c[2]
    # Voltage and temperature parameters:
    if args.scale is not None:
        table.scale = args.scale
    if args.offset is not None:
        table.offset = args.offset
    if args.delay is not None:
        table.delay = args.delay
    if args.unit is not None:
        table.sensor_unit = args.unit
    # NTC parameters:
    if args.ntc_a is not None:
        table.ntc_a = args.ntc_a
    if args.ntc_b is not None:
        table.ntc_b = args.ntc_b
    if args.ntc_c is not None:
        table.ntc_c = args.ntc_c
    if args.ntc_m is not None:
        table.ntc_m = args.ntc_m
    if args.ntc_n is not None:
        table.ntc_n = args.ntc_n
    if args.ntc_k is not None:
        table.ntc_k = args.ntc_k
    # Pulse parameters:
    if args.threshold is not None:
        table.threshold = args.threshold
    if args.hysteresis is not None:
        table.hysteresis = args.hysteresis
    if args.debounce_time is not None:
        table.debounce_time = args.debounce_time
    if args.edge_mask is not None:
        table.edge_mask = args.edge_mask

    table.mfg_info = int(args.manufacturer_info[0])

    if args.debug[0] > 0:
        print('CTid params:\n\t', table)

    try:
        table_data = table.encode(table_version)
    except CTid.Error:
        print('%s: %s' % (parser.prog, sys.exc_info()[1]), file=sys.stderr)
        sys.exit(1)

    if args.debug[0] > 0:
        hexdump('table', bytes((CTid.START_SYM, )) + table_data)

    bitstream = CTid.bitstuff(table_data)

    if args.debug[0] > 0:
        hexdump('bit stream', bitstream)

    ihex = IntelHex(args.program_template[0])

    max_addr = max(ihex.todict().keys())
    if max_addr > CTid_table_addr:
        print('%s: max address in program file 0x%x is >= table address 0x%x'
              % (parser.prog, max_addr, CTid_table_addr))
        sys.exit(1)

    addr = CTid_table_addr
    ihex[addr] = len(bitstream)
    addr += 1
    for i, byte in enumerate(bitstream):
        ihex[addr + i] = byte

    ihex.write_hex_file(args.output_file[0])

parser = None
table = None
if __name__ == '__main__':
    main()



# !!! DO NOT EDIT !!!  Generated by gen-phys-units.
#
#   Released under MIT license.
#   See file LICENSE for details.
#
# pylint: disable=too-many-lines
# pylint: disable=too-few-public-methods

METRIC = 0
IMPERIAL = 1

class UnitTableEntry:
    def __init__(self, fix_scale=None, rate_unit=None,
                 cumul_scale=None, cumul_unit=None,
                 name=None):
        self.fix_scale = fix_scale
        self.rate_unit = rate_unit
        self.cumul_scale = cumul_scale
        self.cumul_unit = cumul_unit
        self.name = name

class ConversionTableEntry:
    def __init__(self, unit=None, time_dependent=False,
                 calc=None, inverse=None):
        self.unit = unit
        self.time_dependent = time_dependent
        self.calc = calc
        self.inverse = inverse

class Units:

    # Preferred units for METRIC and IMPERIAL systems:
    preferred = [
        ['kW', 'kWh', 'kVA', 'kVAh', 'kvar', 'kvahr', 'kg', 'kg/s',
         'kWh/m^2', 'km/h', 'hPa', 'Ah', 'Vh', 'L/m', 'L', '°C',
         '°C·d'],
        ['kW', 'kWh', 'kVA', 'kVAh', 'kvar', 'kvahr', 'lbs', 'lbs/s',
         'kWh/m^2', 'mph', 'inHg', 'Ah', 'Vh', 'ft', '°F', '°F·d',
         'lbs·s', 'inHg·s', 'ft·s', 'gpm', 'gal']
    ]

    table = {
        'P': UnitTableEntry(fix_scale=1, rate_unit='W',
                            cumul_scale=1 / 3600000, cumul_unit='kWh',
                            name='Power'),
        'S': UnitTableEntry(fix_scale=1, rate_unit='VA',
                            cumul_scale=1 / 3600000, cumul_unit='kVAh',
                            name='Apparent Power'),
        'V': UnitTableEntry(fix_scale=1000, rate_unit='V',
                            cumul_scale=1 / 3600000, cumul_unit='Vh',
                            name='Voltage'),
        'I': UnitTableEntry(fix_scale=1000, rate_unit='A',
                            cumul_scale=1 / 3600000, cumul_unit='Ah',
                            name='Current'),
        'F': UnitTableEntry(fix_scale=1000, rate_unit='Hz',
                            cumul_scale=1 / 1000, cumul_unit='Hz·s',
                            name='Frequency'),
        'THD': UnitTableEntry(fix_scale=1000, rate_unit='%',
                              cumul_scale=1 / 1000, cumul_unit='%·s',
                              name='THD'),
        'T': UnitTableEntry(fix_scale=1000, rate_unit='°C',
                            cumul_scale=1 / 1000, cumul_unit='°C·s',
                            name='Temperature'),
        'Q': UnitTableEntry(fix_scale=1, rate_unit='g/s',
                            cumul_scale=1 / 1000, cumul_unit='kg',
                            name='Mass flow'),
        'v': UnitTableEntry(fix_scale=1000, rate_unit='m/s',
                            cumul_scale=1 / 1000, cumul_unit='m',
                            name='Speed'),
        '#': UnitTableEntry(fix_scale=1, rate_unit='',
                            cumul_scale=1, cumul_unit='s',
                            name='Whole number'),
        'R': UnitTableEntry(fix_scale=1, rate_unit='Ω',
                            cumul_scale=1, cumul_unit='Ω·s',
                            name='Resistance'),
        'Ee': UnitTableEntry(fix_scale=1, rate_unit='W/m^2',
                             cumul_scale=1 / 3600000, cumul_unit='kWh/m^2',
                             name='Irradiance'),
        'PQ': UnitTableEntry(fix_scale=1, rate_unit='var',
                             cumul_scale=1 / 3600000, cumul_unit='kvahr',
                             name='Reactive Power'),
        '$': UnitTableEntry(fix_scale=536870912, rate_unit='${currency}/s',
                            cumul_scale=1 / 536870912, cumul_unit='${currency}',
                            name='Monetary'),
        'a': UnitTableEntry(fix_scale=1000, rate_unit='°',
                            cumul_scale=1 / 1000, cumul_unit='°·s',
                            name='Angle'),
        'h': UnitTableEntry(fix_scale=1000, rate_unit='%',
                            cumul_scale=1 / 1000, cumul_unit='%·s',
                            name='Humidity'),
        'Qv': UnitTableEntry(fix_scale=1000000000, rate_unit='m^3/s',
                             cumul_scale=1 / 1000000, cumul_unit='L',
                             name='Volumetric flow'),
        'Pa': UnitTableEntry(fix_scale=1, rate_unit='Pa',
                             cumul_scale=1, cumul_unit='Pa·s',
                             name='Pressure'),
        'Qe': UnitTableEntry(fix_scale=1000, rate_unit='Ah',
                             cumul_scale=1 / 1000, cumul_unit='Ah·s',
                             name='Charge'),
        'Vdc': UnitTableEntry(fix_scale=1000, rate_unit='V',
                              cumul_scale=1 / 3600000, cumul_unit='Vh',
                              name='Voltage (mean)'),
        'Idc': UnitTableEntry(fix_scale=1000, rate_unit='A',
                              cumul_scale=1 / 3600000, cumul_unit='Ah',
                              name='Current (mean)'),
        'm': UnitTableEntry(fix_scale=1000, rate_unit='g',
                            cumul_scale=1 / 1000, cumul_unit='g·s',
                            name='Mass'),
        '#3': UnitTableEntry(fix_scale=1000, rate_unit='',
                             cumul_scale=1 / 1000, cumul_unit='s',
                             name='Number with 3 decimals'),
        'aq': UnitTableEntry(fix_scale=1000, rate_unit='',
                             cumul_scale=1 / 1000, cumul_unit='s',
                             name='Air Quality Index (0=good, 500=bad)'),
        '%': UnitTableEntry(fix_scale=1000, rate_unit='%',
                            cumul_scale=1 / 1000, cumul_unit='%·s',
                            name='Percentage'),
        'ppm': UnitTableEntry(fix_scale=1000, rate_unit='ppm',
                              cumul_scale=1 / 1000, cumul_unit='ppm·s',
                              name='Parts per million'),
        'd': UnitTableEntry(fix_scale=1, rate_unit='',
                            cumul_scale=1, cumul_unit='',
                            name='Discrete number'),
        'dT': UnitTableEntry(fix_scale=1000, rate_unit='°C',
                             cumul_scale=1 / 1000, cumul_unit='°C·s',
                             name='Temperature difference'),
    }

    alternates = {
        'A·s': [
            ConversionTableEntry(
                unit='Ah', time_dependent=False,
                calc=lambda x, t=None: (1/3600)*x,
                inverse=lambda x, t=None: 3600*x),
        ],
        'Pa': [
            ConversionTableEntry(
                unit='hPa', time_dependent=False,
                calc=lambda x, t=None: (1/100)*x,
                inverse=lambda x, t=None: 100*x),
            ConversionTableEntry(
                unit='bar', time_dependent=False,
                calc=lambda x, t=None: 1.0e-5*x,
                inverse=lambda x, t=None: 100000.0*x),
            ConversionTableEntry(
                unit='psi', time_dependent=False,
                calc=lambda x, t=None: 0.00014503773773*x,
                inverse=lambda x, t=None: 6894.7572931783*x),
            ConversionTableEntry(
                unit='inHg', time_dependent=False,
                calc=lambda x, t=None: 0.000295300999981497*x,
                inverse=lambda x, t=None: 3386.375258*x),
        ],
        'Pa·s': [
            ConversionTableEntry(
                unit='psi·s', time_dependent=False,
                calc=lambda x, t=None: 0.00014503773773*x,
                inverse=lambda x, t=None: 6894.7572931783*x),
            ConversionTableEntry(
                unit='inHg·s', time_dependent=False,
                calc=lambda x, t=None: 0.000295300999981497*x,
                inverse=lambda x, t=None: 3386.375258*x),
        ],
        'VA·s': [
            ConversionTableEntry(
                unit='VAh', time_dependent=False,
                calc=lambda x, t=None: (1/3600)*x,
                inverse=lambda x, t=None: 3600*x),
        ],
        'V·s': [
            ConversionTableEntry(
                unit='Vh', time_dependent=False,
                calc=lambda x, t=None: (1/3600)*x,
                inverse=lambda x, t=None: 3600*x),
        ],
        'W': [
            ConversionTableEntry(
                unit='hp', time_dependent=False,
                calc=lambda x, t=None: 0.00135962115516133*x,
                inverse=lambda x, t=None: 735.499*x),
            ConversionTableEntry(
                unit='Btu/h', time_dependent=False,
                calc=lambda x, t=None: 3.412141633*x,
                inverse=lambda x, t=None: 0.293071070183211*x),
        ],
        'W·s': [
            ConversionTableEntry(
                unit='Wh', time_dependent=False,
                calc=lambda x, t=None: (1/3600)*x,
                inverse=lambda x, t=None: 3600*x),
            ConversionTableEntry(
                unit='J', time_dependent=False,
                calc=lambda x, t=None: x,
                inverse=lambda x, t=None: x),
            ConversionTableEntry(
                unit='Btu', time_dependent=False,
                calc=lambda x, t=None: 0.000947817077749151*x,
                inverse=lambda x, t=None: 1055.0559*x),
        ],
        'W·s/m^2': [
            ConversionTableEntry(
                unit='Wh/m^2', time_dependent=False,
                calc=lambda x, t=None: (1/3600)*x,
                inverse=lambda x, t=None: 3600*x),
        ],
        'g': [
            ConversionTableEntry(
                unit='lbs', time_dependent=False,
                calc=lambda x, t=None: 0.00220462*x,
                inverse=lambda x, t=None: 453.59290943564*x),
            ConversionTableEntry(
                unit='T', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
        ],
        'g/s': [
            ConversionTableEntry(
                unit='lbs/s', time_dependent=False,
                calc=lambda x, t=None: 0.00220462*x,
                inverse=lambda x, t=None: 453.59290943564*x),
        ],
        'g·s': [
            ConversionTableEntry(
                unit='lbs·s', time_dependent=False,
                calc=lambda x, t=None: 0.00220462*x,
                inverse=lambda x, t=None: 453.59290943564*x),
        ],
        'm': [
            ConversionTableEntry(
                unit='yd', time_dependent=False,
                calc=lambda x, t=None: 1.09361329833771*x,
                inverse=lambda x, t=None: 0.9144*x),
            ConversionTableEntry(
                unit='ft', time_dependent=False,
                calc=lambda x, t=None: 3.28083989501312*x,
                inverse=lambda x, t=None: 0.3048*x),
            ConversionTableEntry(
                unit='mi', time_dependent=False,
                calc=lambda x, t=None: 0.000621371192237334*x,
                inverse=lambda x, t=None: 1609.344*x),
        ],
        'm/s': [
            ConversionTableEntry(
                unit='m/h', time_dependent=False,
                calc=lambda x, t=None: 3600*x,
                inverse=lambda x, t=None: (1/3600)*x),
            ConversionTableEntry(
                unit='mph', time_dependent=False,
                calc=lambda x, t=None: 2.2369362920544*x,
                inverse=lambda x, t=None: 0.44704*x),
        ],
        'm^3': [
            ConversionTableEntry(
                unit='L', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='gal', time_dependent=False,
                calc=lambda x, t=None: 264.200792602378*x,
                inverse=lambda x, t=None: 0.003785*x),
        ],
        'm^3/s': [
            ConversionTableEntry(
                unit='L/s', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='L/m', time_dependent=False,
                calc=lambda x, t=None: 60000.0*x,
                inverse=lambda x, t=None: 1.66666666666667e-5*x),
            ConversionTableEntry(
                unit='L/h', time_dependent=False,
                calc=lambda x, t=None: 3600000.0*x,
                inverse=lambda x, t=None: 2.77777777777778e-7*x),
            ConversionTableEntry(
                unit='gal/s', time_dependent=False,
                calc=lambda x, t=None: 264.200792602378*x,
                inverse=lambda x, t=None: 0.003785*x),
            ConversionTableEntry(
                unit='gpm', time_dependent=False,
                calc=lambda x, t=None: 15852.0475561427*x,
                inverse=lambda x, t=None: 6.30833333333334e-5*x),
            ConversionTableEntry(
                unit='gph', time_dependent=False,
                calc=lambda x, t=None: 951122.85336856*x,
                inverse=lambda x, t=None: 1.05138888888889e-6*x),
        ],
        'm·s': [
            ConversionTableEntry(
                unit='ft·s', time_dependent=False,
                calc=lambda x, t=None: 3.28083989501312*x,
                inverse=lambda x, t=None: 0.3048*x),
        ],
        'var·s': [
            ConversionTableEntry(
                unit='vahr', time_dependent=False,
                calc=lambda x, t=None: (1/3600)*x,
                inverse=lambda x, t=None: 3600*x),
        ],
        '°C': [
            ConversionTableEntry(
                unit='°F', time_dependent=False,
                calc=lambda x, t=None: (9/5)*x + 32,
                inverse=lambda x, t=None: (5/9)*x - 160/9),
            ConversionTableEntry(
                unit='°K', time_dependent=False,
                calc=lambda x, t=None: x + 273.15,
                inverse=lambda x, t=None: x - 273.15),
        ],
        '°C·s': [
            ConversionTableEntry(
                unit='°K·s', time_dependent=True,
                calc=lambda x, t=None: 273.15*t + x,
                inverse=lambda x, t=None: -273.15*t + x),
            ConversionTableEntry(
                unit='°F·s', time_dependent=True,
                calc=lambda x, t=None: 32*t + (9/5)*x,
                inverse=lambda x, t=None: -160/9*t + (5/9)*x),
        ]
    }

    scaled = {
        'A': [
            ConversionTableEntry(
                unit='yA', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zA', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aA', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fA', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pA', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nA', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μA', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mA', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kA', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'Ah': [
            ConversionTableEntry(
                unit='yAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pAh', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nAh', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μAh', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mAh', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kAh', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'Btu/h': [
            ConversionTableEntry(
                unit='MBH', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
        ],
        'Hz': [
            ConversionTableEntry(
                unit='yHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pHz', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nHz', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μHz', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mHz', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kHz', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='THz', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YHz', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'T': [
            ConversionTableEntry(
                unit='kT', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MT', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GT', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TT', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PT', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
        ],
        'V': [
            ConversionTableEntry(
                unit='yV', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zV', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aV', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fV', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pV', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nV', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μV', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mV', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kV', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MV', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GV', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TV', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PV', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EV', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZV', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YV', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'VA': [
            ConversionTableEntry(
                unit='yVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pVA', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nVA', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μVA', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mVA', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kVA', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YVA', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'VAh': [
            ConversionTableEntry(
                unit='yVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pVAh', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nVAh', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μVAh', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mVAh', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kVAh', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YVAh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'Vh': [
            ConversionTableEntry(
                unit='yVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pVh', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nVh', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μVh', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mVh', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kVh', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YVh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'W': [
            ConversionTableEntry(
                unit='yW', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zW', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aW', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fW', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pW', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nW', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μW', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mW', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kW', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MW', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GW', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TW', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PW', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EW', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZW', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YW', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'W/m^2': [
            ConversionTableEntry(
                unit='yW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kW/m^2', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YW/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'Wh': [
            ConversionTableEntry(
                unit='yWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pWh', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nWh', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μWh', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mWh', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kWh', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YWh', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'Wh/m^2': [
            ConversionTableEntry(
                unit='yWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YWh/m^2', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'g': [
            ConversionTableEntry(
                unit='μg', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mg', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kg', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
        ],
        'g/s': [
            ConversionTableEntry(
                unit='μg/s', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mg/s', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kg/s', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='T/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='kT/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='MT/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='GT/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='TT/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='PT/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
        ],
        'g·s': [
            ConversionTableEntry(
                unit='μg·s', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mg·s', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kg·s', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='T·s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='kT·s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='MT·s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='GT·s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='TT·s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='PT·s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
        ],
        'm': [
            ConversionTableEntry(
                unit='ym', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zm', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='am', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fm', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pm', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nm', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μm', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mm', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='km', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='Mm', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='Gm', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='Tm', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='Pm', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='Em', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='Zm', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='Ym', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'm/h': [
            ConversionTableEntry(
                unit='ym/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zm/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='am/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fm/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pm/h', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nm/h', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μm/h', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mm/h', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='km/h', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='Mm/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='Gm/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='Tm/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='Pm/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='Em/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='Zm/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='Ym/h', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'm/s': [
            ConversionTableEntry(
                unit='ym/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zm/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='am/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fm/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pm/s', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nm/s', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μm/s', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mm/s', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='km/s', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='Mm/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='Gm/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='Tm/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='Pm/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='Em/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='Zm/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='Ym/s', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'vahr': [
            ConversionTableEntry(
                unit='yvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='avahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pvahr', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nvahr', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μvahr', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mvahr', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kvahr', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='Mvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='Gvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='Tvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='Pvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='Evahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='Zvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='Yvahr', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        'var': [
            ConversionTableEntry(
                unit='yvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='avar', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pvar', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nvar', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μvar', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mvar', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kvar', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='Mvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='Gvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='Tvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='Pvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='Evar', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='Zvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='Yvar', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ],
        '°C·s': [
            ConversionTableEntry(
                unit='°C·h', time_dependent=False,
                calc=lambda x, t=None: 0.000277777777777778*x,
                inverse=lambda x, t=None: 3600.0*x),
            ConversionTableEntry(
                unit='°C·d', time_dependent=False,
                calc=lambda x, t=None: 1.15740740740741e-5*x,
                inverse=lambda x, t=None: 86399.9999999998*x),
        ],
        '°F·s': [
            ConversionTableEntry(
                unit='°F·h', time_dependent=False,
                calc=lambda x, t=None: 0.000277777777777778*x,
                inverse=lambda x, t=None: 3600.0*x),
            ConversionTableEntry(
                unit='°F·d', time_dependent=False,
                calc=lambda x, t=None: 1.15740740740741e-5*x,
                inverse=lambda x, t=None: 86399.9999999998*x),
        ],
        '°K·s': [
            ConversionTableEntry(
                unit='°K·h', time_dependent=False,
                calc=lambda x, t=None: 0.000277777777777778*x,
                inverse=lambda x, t=None: 3600.0*x),
            ConversionTableEntry(
                unit='°K·d', time_dependent=False,
                calc=lambda x, t=None: 1.15740740740741e-5*x,
                inverse=lambda x, t=None: 86399.9999999998*x),
        ],
        'Ω': [
            ConversionTableEntry(
                unit='yΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e+24*x,
                inverse=lambda x, t=None: 1.0e-24*x),
            ConversionTableEntry(
                unit='zΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e+21*x,
                inverse=lambda x, t=None: 1.0e-21*x),
            ConversionTableEntry(
                unit='aΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e+18*x,
                inverse=lambda x, t=None: 1.0e-18*x),
            ConversionTableEntry(
                unit='fΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e+15*x,
                inverse=lambda x, t=None: 1.0e-15*x),
            ConversionTableEntry(
                unit='pΩ', time_dependent=False,
                calc=lambda x, t=None: 1000000000000.0*x,
                inverse=lambda x, t=None: 1.0e-12*x),
            ConversionTableEntry(
                unit='nΩ', time_dependent=False,
                calc=lambda x, t=None: 1000000000.0*x,
                inverse=lambda x, t=None: 1.0e-9*x),
            ConversionTableEntry(
                unit='μΩ', time_dependent=False,
                calc=lambda x, t=None: 1000000.0*x,
                inverse=lambda x, t=None: 1.0e-6*x),
            ConversionTableEntry(
                unit='mΩ', time_dependent=False,
                calc=lambda x, t=None: 1000.0*x,
                inverse=lambda x, t=None: 0.001*x),
            ConversionTableEntry(
                unit='kΩ', time_dependent=False,
                calc=lambda x, t=None: 0.001*x,
                inverse=lambda x, t=None: 1000.0*x),
            ConversionTableEntry(
                unit='MΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e-6*x,
                inverse=lambda x, t=None: 1000000.0*x),
            ConversionTableEntry(
                unit='GΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e-9*x,
                inverse=lambda x, t=None: 1000000000.0*x),
            ConversionTableEntry(
                unit='TΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e-12*x,
                inverse=lambda x, t=None: 1000000000000.0*x),
            ConversionTableEntry(
                unit='PΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e-15*x,
                inverse=lambda x, t=None: 1.0e+15*x),
            ConversionTableEntry(
                unit='EΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e-18*x,
                inverse=lambda x, t=None: 1.0e+18*x),
            ConversionTableEntry(
                unit='ZΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e-21*x,
                inverse=lambda x, t=None: 1.0e+21*x),
            ConversionTableEntry(
                unit='YΩ', time_dependent=False,
                calc=lambda x, t=None: 1.0e-24*x,
                inverse=lambda x, t=None: 1.0e+24*x),
        ]
    }

    @classmethod
    def set_currency(cls, currency):
        Units.table['$'].rate_unit = f'{currency}/s'
        Units.table['$'].cumul_unit = f'{currency}'



#!/usr/bin/env python3
#
#   Copyright (c) 2021-2022 eGauge Systems LLC
#     1644 Conestoga St, Suite 2
#     Boulder, CO 80301
#     voice: 720-545-9767
#     email: dave@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# pylint: disable=no-name-in-module
#
'''This module provides a helper function to ask for eGauge cloud API
credentials (aka eGuard credentials) via a popup dialog.  To use it,
pass the ask method to the TokenAuth's ask parameter after partially
evaluating it to bind the parent Qt window to "parent":

from functools import partial

import egauge.webapi.cloud.credentials

    # this `self' is a Pyside2 Ui_MainWindow:
    auth = webapi.auth.TokenAuth(ask=partial(credentials.ask, self))
    self.sn_api = webapi.cloud.SerialNumber(auth=auth)

'''
import getpass

from PySide2.QtWidgets import QDialog
from .gui.credentials_dialog import Ui_Credentials_Dialog

class LoginCanceled(Exception):
    '''Raised when the user cancels a login request.'''

class Credentials_Manager:
    def __init__(self, gui_parent=None):
        '''Create a credentials manager.  The task of this manager is mainly
        to track if a previous login failed.  The user of this object
        should set the previous_login_failed member to False after the
        credentials have been used successfully.

        If GUI_PARENT is not None, it must be the QT5 (PySide2) parent
        window to use for the dialog.  If it is None, the credentials
        will be requested via standard I/O (getpass).

        '''
        self.parent = gui_parent
        self.previous_login_failed = False

    def ask(self):
        '''Ask for the username, password, and optional token for an eGauge
        cloud API account (eGuard account).  Returns a tuple
        containing a username and password or raises LoginCanceled if
        the user presses the "Cancel" button.

        '''
        if self.parent:
            dialog = Credentials_Dialog(self.parent,
                                        self.previous_login_failed)
            dialog.exec_()
            if not dialog.accepted:
                raise LoginCanceled()
            self.previous_login_failed = True
            return (dialog.username, dialog.password + dialog.token)

        fail_msg = ''
        if self.previous_login_failed:
            fail_msg = 'Login failed.  '
        print('%sPlease enter eGuard credentials.' % fail_msg)
        try:
            usr = input('Username: ')
            pwd = getpass.getpass(prompt='Password[+token]: ')
        except (KeyboardInterrupt, EOFError) as e:
            raise LoginCanceled from e
        self.previous_login_failed = True
        return [usr, pwd]

class Credentials_Dialog(QDialog, Ui_Credentials_Dialog):
    def __init__(self, parent, failed):
        self.accepted = False
        self.username = None
        self.password = None
        self.token = ''
        super().__init__(parent)
        self.setupUi(self)
        if failed:
            prompt = 'Login failed. ' + self.prompt_label.text()
            self.prompt_label.setText(prompt)
        self.username_lineEdit.setFocus()

    def exec_(self):
        self.accepted = False
        super().exec_()

    def accept(self):
        super().accept()
        self.accepted = True
        self.username = self.username_lineEdit.text()
        self.password = self.password_lineEdit.text()
        self.token = self.token_lineEdit.text()



import os
import setuptools

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setuptools.setup(
    name='egauge-python',
    version='0.5.6',
    packages=setuptools.find_namespace_packages(include='egauge.*'),
    install_requires=['crcmod', 'intelhex', 'wheel', 'pexpect',
                      'requests>=2.4.2'],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'ctid-encoder = egauge.ctid.encoder:main'
        ]
    },
    license='MIT License',  # example license
    description='.',
    long_description=README,
    long_description_content_type='text/x-rst',
    url='https://bitbucket.org/egauge/python/',
    author='David Mosberger-Tang',
    author_email='davidm@egauge.net',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6'
    ],
)



#
# Copyright (c) 2020-2022 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
'''This module provides helper methods for accessing JSON web services.'''
import logging
import json

import requests

from .error import Error

log = logging.getLogger(__name__)

class JSONAPIError(Error):
    '''Raised if for any JSON API errors.  The first argument to this
    exception is the 401 response received from the web server.

    '''

class UnauthenticatedError(Error):
    '''Raised when a request fails with HTTP status code 401.'''

def get(resource, **kwargs):
    '''Issue GET request for RESOURCE and return the parsed JSON data or
    None if the request failed or returned invalid JSON data.
    Additional keyword arguments are passed on to requests.get().

    '''
    try:
        r = requests.get(resource, **kwargs)
    except requests.exceptions.RequestException as e:
        raise JSONAPIError('requests.get exception.', e) from e
    if r.status_code == 401:
        raise UnauthenticatedError(r)
    if r.status_code < 200 or r.status_code > 299:
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.exception('HTTP GET status code %s.  Keyword args: %s',
                          r.status_code, kwargs)
        raise JSONAPIError('Unexpected HTTP status code.', r.status_code, r.content)
    reply = None
    try:
        if r.text:
            reply = r.json()
    except ValueError as e:
        raise JSONAPIError('Invalid JSON data.', r.content) from e
    return reply

def put(resource, json_data, **kwargs):
    '''Issue PUT request with JSON_DATA as body to RESOURCE and return
    parsed JSON reply or None if the request failed or returned
    invalid JSON data.  Additional keyword arguments are passed on to
    requests.put().

    '''
    headers = kwargs.get('headers', {})
    headers['Content-Type'] = 'application/json'
    kwargs['headers'] = headers
    try:
        r = requests.put(resource, json=json_data, **kwargs)
    except requests.exceptions.RequestException as e:
        raise JSONAPIError('requests.put exception.', e) from e
    if r.status_code == 401:
        raise UnauthenticatedError(r)
    if not 200 <= r.status_code <= 299:
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.exception('HTTP PUT status code %s.  '
                          'Resource %s, Data: %s, keyword args: %s',
                          r.status_code, resource, json_data, kwargs)
        raise JSONAPIError('Unexpected HTTP status code.', r.status_code, r.content)
    reply = None
    try:
        if r.text:
            reply = r.json()
    except ValueError as e:
        raise JSONAPIError('Invalid JSON data.', r.content) from e
    return reply

def post(resource, json_data, **kwargs):
    '''Issue POST request with JSON_DATA as body to RESOURCE and return
    parsed JSON reply or None if the request failed or returned
    invalid JSON data.  Additional keyword arguments are passed on to
    requests.post().

    '''
    headers = kwargs.get('headers', {})
    headers['Content-Type'] = 'application/json'
    kwargs['headers'] = headers
    try:
        r = requests.post(resource, json=json_data, **kwargs)
    except requests.exceptions.RequestException as e:
        raise JSONAPIError('requests.post exception.', e) from e
    if r.status_code == 401:
        raise UnauthenticatedError(r)
    if not 200 <= r.status_code <= 299:
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.exception('HTTP POST status code %s.  '
                          'Resource %s, Data: %s, keyword args: %s',
                          r.status_code, resource, json_data, kwargs)
        raise JSONAPIError('Unexpected HTTP status code.', r.status_code, r.content)
    reply = None
    try:
        if r.text:
            reply = r.json()
    except ValueError as e:
        raise JSONAPIError('Invalid JSON data.', r.content) from e
    return reply

def delete(resource, **kwargs):
    '''Issue DELETE request for RESOURCE and return the parsed JSON data
    or None if the request failed or returned invalid JSON data.
    Additional keyword arguments are passed on to requests.delete().

    '''
    try:
        r = requests.delete(resource, **kwargs)
    except requests.exceptions.RequestException as e:
        raise JSONAPIError('requests.delete exception.', e) from e
    if r.status_code == 401:
        raise UnauthenticatedError(r)
    if r.status_code < 200 or r.status_code > 299:
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.exception('HTTP DELETE status code %s.  Keyword args: %s',
                          r.status_code, kwargs)
        raise JSONAPIError('Unexpected HTTP status code.', r.status_code, r.content)
    reply = None
    try:
        if r.text:
            reply = r.json()
    except ValueError as e:
        raise JSONAPIError('Invalid JSON data.', r.content) from e
    return reply
